package com.sun.jmx.remote.internal;

import javax.management.QueryEval;
import javax.management.NotificationBroadcaster;
import javax.management.NotificationFilterSupport;
import java.security.PrivilegedActionException;
import java.security.PrivilegedAction;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import javax.management.MBeanServerDelegate;
import javax.management.InstanceNotFoundException;
import java.util.ArrayList;
import javax.management.remote.TargetedNotification;
import javax.management.remote.NotificationResult;
import javax.management.MBeanServerNotification;
import javax.management.Notification;
import java.util.HashSet;
import java.util.Iterator;
import com.sun.jmx.remote.util.EnvHelp;
import java.util.Collections;
import java.util.Map;
import javax.management.ObjectName;
import java.util.Set;
import com.sun.jmx.remote.util.ClassLogger;
import javax.management.NotificationFilter;
import javax.management.QueryExp;
import javax.management.NotificationListener;
import java.util.Collection;
import javax.management.MBeanServer;
import java.util.HashMap;

public class ArrayNotificationBuffer implements NotificationBuffer
{
    private boolean disposed;
    private static final Object globalLock;
    private static final HashMap<MBeanServer, ArrayNotificationBuffer> mbsToBuffer;
    private final Collection<ArrayNotificationBuffer.ShareBuffer> sharers;
    private final NotificationListener bufferListener;
    private static final QueryExp broadcasterQuery;
    private static final NotificationFilter creationFilter;
    private final NotificationListener creationListener;
    private static final ClassLogger logger;
    private final MBeanServer mBeanServer;
    private final ArrayQueue<ArrayNotificationBuffer.NamedNotification> queue;
    private int queueSize;
    private long earliestSequenceNumber;
    private long nextSequenceNumber;
    private Set<ObjectName> createdDuringQuery;
    static final String broadcasterClass;
    
    public static NotificationBuffer getNotificationBuffer(MBeanServer mBeanServer, Map<String, ?> emptyMap) {
        if (emptyMap == null) {
            emptyMap = Collections.emptyMap();
        }
        int notifBufferSize = EnvHelp.getNotifBufferSize(emptyMap);
        ArrayNotificationBuffer value;
        int n;
        Object o;
        synchronized (ArrayNotificationBuffer.globalLock) {
            value = (ArrayNotificationBuffer)ArrayNotificationBuffer.mbsToBuffer.get(mBeanServer);
            n = (value == null);
            if (!(n == 0)) {
                value = new ArrayNotificationBuffer(mBeanServer, notifBufferSize);
                ArrayNotificationBuffer.mbsToBuffer.put(mBeanServer, value);
            }
            value.getClass();
            o = value.new ShareBuffer(notifBufferSize);
        }
        if (!(n == 0)) {
            value.createListeners();
            return o;
        }
        return o;
    }
    
    static void removeNotificationBuffer(MBeanServer key) {
        synchronized (ArrayNotificationBuffer.globalLock) {
            ArrayNotificationBuffer.mbsToBuffer.remove(key);
        }
    }
    
    void addSharer(ArrayNotificationBuffer.ShareBuffer shareBuffer) {
        synchronized (ArrayNotificationBuffer.globalLock) {
            synchronized (this) {
                if (shareBuffer.getSize() > this.queueSize) {
                    this.resize(shareBuffer.getSize());
                }
            }
            this.sharers.add(shareBuffer);
        }
    }
    
    private void removeSharer(ArrayNotificationBuffer.ShareBuffer shareBuffer) {
        int empty;
        synchronized (ArrayNotificationBuffer.globalLock) {
            this.sharers.remove(shareBuffer);
            empty = this.sharers.isEmpty();
            if (!(empty == 0)) {
                removeNotificationBuffer(this.mBeanServer);
            }
            else {
                int n = 0;
                Iterator<ArrayNotificationBuffer.ShareBuffer> iterator = this.sharers.iterator();
                while (iterator.hasNext()) {
                    int size = ((ArrayNotificationBuffer.ShareBuffer)iterator.next()).getSize();
                    if (size > n) {
                        n = size;
                    }
                    else {
                        continue;
                    }
                }
                if (n < this.queueSize) {
                    this.resize(n);
                }
            }
        }
        if (!(empty == 0)) {
            synchronized (this) {
                this.disposed = true;
                this.notifyAll();
            }
            this.destroyListeners();
        }
    }
    
    private synchronized void resize(int queueSize) {
        if (queueSize == this.queueSize) {
            return;
        }
        while (this.queue.size() > queueSize) {
            this.dropNotification();
        }
        this.queue.resize(queueSize);
        this.queueSize = queueSize;
        return;
    }
    
    private ArrayNotificationBuffer(MBeanServer mBeanServer, int n) {
        this.disposed = false;
        this.sharers = new HashSet(1);
        this.bufferListener = new ArrayNotificationBuffer.BufferListener();
        this.creationListener = new NotificationListener() {
            @Override
            public void handleNotification(Notification notification, Object o) {
                ArrayNotificationBuffer.logger.debug("creationListener", "handleNotification called");
                ArrayNotificationBuffer.access$600(ArrayNotificationBuffer.this, (MBeanServerNotification)notification);
            }
        };
        if (ArrayNotificationBuffer.logger.traceOn()) {
            ArrayNotificationBuffer.logger.trace("Constructor", "queueSize=" + n);
        }
        if (mBeanServer == null || n < 1) {
            throw new IllegalArgumentException("Bad args");
        }
        this.mBeanServer = mBeanServer;
        this.queueSize = n;
        this.queue = new ArrayQueue(n);
        this.earliestSequenceNumber = System.currentTimeMillis();
        this.nextSequenceNumber = this.earliestSequenceNumber;
        ArrayNotificationBuffer.logger.trace("Constructor", "ends");
        return;
    }
    
    private synchronized boolean isDisposed() {
        return this.disposed;
    }
    
    @Override
    public void dispose() {
        throw new UnsupportedOperationException();
    }
    
    @Override
    public NotificationResult fetchNotifications(NotificationBufferFilter obj, long lng, long lng2, int i) throws InterruptedException {
        ArrayNotificationBuffer.logger.trace("fetchNotifications", "starts");
        if (lng < 0L || this.isDisposed()) {
            synchronized (this) {
                return new NotificationResult(this.earliestSequenceNumber(), this.nextSequenceNumber(), new TargetedNotification[0]);
            }
        }
        if (obj == null || lng < 0L || lng2 < 0L || i < 0) {
            ArrayNotificationBuffer.logger.trace("fetchNotifications", "Bad args");
            throw new IllegalArgumentException("Bad args to fetch");
        }
        if (ArrayNotificationBuffer.logger.debugOn()) {
            ArrayNotificationBuffer.logger.trace("fetchNotifications", "filter=" + obj + "; startSeq=" + lng + "; timeout=" + lng2 + "; max=" + i);
        }
        if (lng > this.nextSequenceNumber()) {
            String string = "Start sequence number too big: " + lng + " > " + this.nextSequenceNumber();
            ArrayNotificationBuffer.logger.trace("fetchNotifications", string);
            throw new IllegalArgumentException(string);
        }
        long lng3 = System.currentTimeMillis() + lng2;
        if (lng3 < 0L) {
            lng3 = Long.MAX_VALUE;
        }
        if (ArrayNotificationBuffer.logger.debugOn()) {
            ArrayNotificationBuffer.logger.debug("fetchNotifications", "endTime=" + lng3);
        }
        long earliestSequenceNumber = -1L;
        ArrayList o = new ArrayList();
        while (true) {
            ArrayNotificationBuffer.logger.debug("fetchNotifications", "main loop starts");
            ArrayNotificationBuffer.NamedNotification notification;
            synchronized (this) {
                if (earliestSequenceNumber < 0L) {
                    earliestSequenceNumber = this.earliestSequenceNumber();
                    if (ArrayNotificationBuffer.logger.debugOn()) {
                        ArrayNotificationBuffer.logger.debug("fetchNotifications", "earliestSeq=" + earliestSequenceNumber);
                    }
                    if (lng < earliestSequenceNumber) {
                        lng = earliestSequenceNumber;
                        ArrayNotificationBuffer.logger.debug("fetchNotifications", "nextSeq=earliestSeq");
                    }
                }
                else {
                    earliestSequenceNumber = this.earliestSequenceNumber();
                }
                if (lng < earliestSequenceNumber) {
                    ArrayNotificationBuffer.logger.trace("fetchNotifications", "nextSeq=" + lng + " < earliestSeq=" + earliestSequenceNumber + " so may have lost notifs");
                    break;
                }
                if (lng < this.nextSequenceNumber()) {
                    notification = this.notificationAt(lng);
                    if (!(obj instanceof ServerNotifForwarder.NotifForwarderBufferFilter)) {
                        try {
                            ServerNotifForwarder.checkMBeanPermission(this.mBeanServer, notification.getObjectName(), "addNotificationListener");
                        }
                        catch (InstanceNotFoundException | SecurityException obj2) {
                            if (ArrayNotificationBuffer.logger.debugOn()) {
                                ArrayNotificationBuffer.logger.debug("fetchNotifications", "candidate: " + notification + " skipped. exception " + obj2);
                            }
                            ++lng;
                            continue;
                        }
                    }
                    if (ArrayNotificationBuffer.logger.debugOn()) {
                        ArrayNotificationBuffer.logger.debug("fetchNotifications", "candidate: " + notification);
                        ArrayNotificationBuffer.logger.debug("fetchNotifications", "nextSeq now " + lng);
                    }
                }
                else {
                    if (o.size() > 0) {
                        ArrayNotificationBuffer.logger.debug("fetchNotifications", "no more notifs but have some so don't wait");
                        break;
                    }
                    long n = lng3 - System.currentTimeMillis();
                    if (n <= 0L) {
                        ArrayNotificationBuffer.logger.debug("fetchNotifications", "timeout");
                        break;
                    }
                    if (this.isDisposed()) {
                        if (ArrayNotificationBuffer.logger.debugOn()) {
                            ArrayNotificationBuffer.logger.debug("fetchNotifications", "dispose callled, no wait");
                        }
                        return new NotificationResult(this.earliestSequenceNumber(), this.nextSequenceNumber(), new TargetedNotification[0]);
                    }
                    else {
                        if (ArrayNotificationBuffer.logger.debugOn()) {
                            ArrayNotificationBuffer.logger.debug("fetchNotifications", "wait(" + n + ")");
                        }
                        this.wait(n);
                        continue;
                    }
                }
            }
            ObjectName objectName = notification.getObjectName();
            Notification notification2 = notification.getNotification();
            ArrayList obj3 = new ArrayList();
            ArrayNotificationBuffer.logger.debug("fetchNotifications", "applying filter to candidate");
            obj.apply(obj3, objectName, notification2);
            if (obj3.size() > 0) {
                if (i <= 0) {
                    ArrayNotificationBuffer.logger.debug("fetchNotifications", "reached maxNotifications");
                    break;
                }
                --i;
                if (ArrayNotificationBuffer.logger.debugOn()) {
                    ArrayNotificationBuffer.logger.debug("fetchNotifications", "add: " + obj3);
                }
                o.addAll(obj3);
            }
            ++lng;
        }
        TargetedNotification[] targetedNotifications = new TargetedNotification[o.size()];
        o.toArray(targetedNotifications);
        NotificationResult o2 = new NotificationResult(earliestSequenceNumber, lng, targetedNotifications);
        if (ArrayNotificationBuffer.logger.debugOn()) {
            ArrayNotificationBuffer.logger.debug("fetchNotifications", o2.toString());
        }
        ArrayNotificationBuffer.logger.trace("fetchNotifications", "ends");
        return o2;
    }
    
    synchronized long earliestSequenceNumber() {
        return this.earliestSequenceNumber;
    }
    
    synchronized long nextSequenceNumber() {
        return this.nextSequenceNumber;
    }
    
    synchronized void addNotification(ArrayNotificationBuffer.NamedNotification namedNotification) {
        if (ArrayNotificationBuffer.logger.traceOn()) {
            ArrayNotificationBuffer.logger.trace("addNotification", namedNotification.toString());
        }
        while (this.queue.size() >= this.queueSize) {
            this.dropNotification();
            if (!ArrayNotificationBuffer.logger.debugOn()) {
                continue;
            }
            ArrayNotificationBuffer.logger.debug("addNotification", "dropped oldest notif, earliestSeq=" + this.earliestSequenceNumber);
        }
        this.queue.add(namedNotification);
        ++this.nextSequenceNumber;
        if (ArrayNotificationBuffer.logger.debugOn()) {
            ArrayNotificationBuffer.logger.debug("addNotification", "nextSeq=" + this.nextSequenceNumber);
        }
        this.notifyAll();
    }
    
    private void dropNotification() {
        this.queue.remove(0);
        ++this.earliestSequenceNumber;
    }
    
    synchronized ArrayNotificationBuffer.NamedNotification notificationAt(long lng) {
        long n = lng - this.earliestSequenceNumber;
        if (n < 0L || n > 2147483647L) {
            String string = "Bad sequence number: " + lng + " (earliest " + this.earliestSequenceNumber + ")";
            ArrayNotificationBuffer.logger.trace("notificationAt", string);
            throw new IllegalArgumentException(string);
        }
        return (ArrayNotificationBuffer.NamedNotification)this.queue.get((int)n);
    }
    
    private void createListeners() {
        ArrayNotificationBuffer.logger.debug("createListeners", "starts");
        synchronized (this) {
            this.createdDuringQuery = new HashSet();
        }
        try {
            this.addNotificationListener(MBeanServerDelegate.DELEGATE_NAME, this.creationListener, ArrayNotificationBuffer.creationFilter, null);
            ArrayNotificationBuffer.logger.debug("createListeners", "added creationListener");
        }
        catch (Exception ex) {
            IllegalArgumentException throwable = new IllegalArgumentException("Can't add listener to MBean server delegate: " + ex);
            EnvHelp.initCause(throwable, ex);
            ArrayNotificationBuffer.logger.fine("createListeners", "Can't add listener to MBean server delegate: " + ex);
            ArrayNotificationBuffer.logger.debug("createListeners", ex);
            throw throwable;
        }
        HashSet set = new HashSet(this.queryNames(null, ArrayNotificationBuffer.broadcasterQuery));
        synchronized (this) {
            set.addAll(this.createdDuringQuery);
            this.createdDuringQuery = null;
        }
        Iterator<Object> iterator = set.iterator();
        while (iterator.hasNext()) {
            this.addBufferListener((ObjectName)iterator.next());
        }
        ArrayNotificationBuffer.logger.debug("createListeners", "ends");
    }
    
    private void addBufferListener(ObjectName objectName) {
        this.checkNoLocks();
        if (ArrayNotificationBuffer.logger.debugOn()) {
            ArrayNotificationBuffer.logger.debug("addBufferListener", objectName.toString());
        }
        try {
            this.addNotificationListener(objectName, this.bufferListener, null, objectName);
        }
        catch (Exception t) {
            ArrayNotificationBuffer.logger.trace("addBufferListener", t);
        }
    }
    
    private void removeBufferListener(ObjectName objectName) {
        this.checkNoLocks();
        if (ArrayNotificationBuffer.logger.debugOn()) {
            ArrayNotificationBuffer.logger.debug("removeBufferListener", objectName.toString());
        }
        try {
            this.removeNotificationListener(objectName, this.bufferListener);
        }
        catch (Exception t) {
            ArrayNotificationBuffer.logger.trace("removeBufferListener", t);
        }
    }
    
    private void addNotificationListener(final ObjectName objectName, final NotificationListener notificationListener, final NotificationFilter notificationFilter, final Object o) throws Exception {
        try {
            AccessController.doPrivileged(new PrivilegedExceptionAction<Void>() {
                @Override
                public Void run() throws InstanceNotFoundException {
                    ArrayNotificationBuffer.this.mBeanServer.addNotificationListener(objectName, notificationListener, notificationFilter, o);
                    return null;
                }
            });
        }
        catch (Exception ex) {
            throw extractException(ex);
        }
    }
    
    private void removeNotificationListener(final ObjectName objectName, final NotificationListener notificationListener) throws Exception {
        try {
            AccessController.doPrivileged(new PrivilegedExceptionAction<Void>() {
                @Override
                public Void run() throws Exception {
                    ArrayNotificationBuffer.this.mBeanServer.removeNotificationListener(objectName, notificationListener);
                    return null;
                }
            });
        }
        catch (Exception ex) {
            throw extractException(ex);
        }
    }
    
    private Set<ObjectName> queryNames(final ObjectName objectName, final QueryExp queryExp) {
        PrivilegedAction<Set<ObjectName>> action = new PrivilegedAction<Set<ObjectName>>() {
            @Override
            public Set<ObjectName> run() {
                return ArrayNotificationBuffer.this.mBeanServer.queryNames(objectName, queryExp);
            }
        };
        try {
            return (Set)AccessController.doPrivileged(action);
        }
        catch (RuntimeException ex) {
            ArrayNotificationBuffer.logger.fine("queryNames", "Failed to query names: " + ex);
            ArrayNotificationBuffer.logger.debug("queryNames", ex);
            throw ex;
        }
    }
    
    private static boolean isInstanceOf(final MBeanServer mBeanServer, final ObjectName objectName, final String s) {
        PrivilegedExceptionAction<Boolean> action = new PrivilegedExceptionAction<Boolean>() {
            @Override
            public Boolean run() throws InstanceNotFoundException {
                return Boolean.valueOf(mBeanServer.isInstanceOf(objectName, s));
            }
        };
        try {
            return (boolean)AccessController.doPrivileged(action);
        }
        catch (Exception ex) {
            ArrayNotificationBuffer.logger.fine("isInstanceOf", "failed: " + ex);
            ArrayNotificationBuffer.logger.debug("isInstanceOf", ex);
            return false;
        }
    }
    
    private void createdNotification(MBeanServerNotification mBeanServerNotification) {
        if (!mBeanServerNotification.getType().equals("JMX.mbean.registered")) {
            ArrayNotificationBuffer.logger.warning("createNotification", "bad type: " + mBeanServerNotification.getType());
            return;
        }
        ObjectName mBeanName = mBeanServerNotification.getMBeanName();
        if (ArrayNotificationBuffer.logger.debugOn()) {
            ArrayNotificationBuffer.logger.debug("createdNotification", "for: " + mBeanName);
        }
        synchronized (this) {
            if (this.createdDuringQuery != null) {
                this.createdDuringQuery.add(mBeanName);
                return;
            }
        }
        if (isInstanceOf(this.mBeanServer, mBeanName, ArrayNotificationBuffer.broadcasterClass)) {
            this.addBufferListener(mBeanName);
            if (this.isDisposed()) {
                this.removeBufferListener(mBeanName);
            }
        }
        return;
    }
    
    private void destroyListeners() {
        this.checkNoLocks();
        ArrayNotificationBuffer.logger.debug("destroyListeners", "starts");
        try {
            this.removeNotificationListener(MBeanServerDelegate.DELEGATE_NAME, this.creationListener);
        }
        catch (Exception t) {
            ArrayNotificationBuffer.logger.warning("remove listener from MBeanServer delegate", t);
        }
        for (ObjectName obj : this.queryNames(null, ArrayNotificationBuffer.broadcasterQuery)) {
            if (ArrayNotificationBuffer.logger.debugOn()) {
                ArrayNotificationBuffer.logger.debug("destroyListeners", "remove listener from " + obj);
            }
            this.removeBufferListener(obj);
        }
        ArrayNotificationBuffer.logger.debug("destroyListeners", "ends");
    }
    
    private void checkNoLocks() {
        if (Thread.holdsLock(this) || Thread.holdsLock(ArrayNotificationBuffer.globalLock)) {
            ArrayNotificationBuffer.logger.warning("checkNoLocks", "lock protocol violation");
        }
    }
    
    private static Exception extractException(Exception exception) {
        while (exception instanceof PrivilegedActionException) {
            exception = ((PrivilegedActionException)exception).getException();
        }
        return exception;
    }
    
    static void access$000(ArrayNotificationBuffer arrayNotificationBuffer, ArrayNotificationBuffer.ShareBuffer shareBuffer) {
        arrayNotificationBuffer.removeSharer(shareBuffer);
    }
    
    static boolean access$400(MBeanServer mBeanServer, ObjectName objectName, String s) {
        return isInstanceOf(mBeanServer, objectName, s);
    }
    
    static void access$600(ArrayNotificationBuffer arrayNotificationBuffer, MBeanServerNotification mBeanServerNotification) {
        arrayNotificationBuffer.createdNotification(mBeanServerNotification);
    }
    
    static {
        globalLock = new Object();
        mbsToBuffer = new HashMap(1);
        broadcasterQuery = new ArrayNotificationBuffer.BroadcasterQuery();
        NotificationFilterSupport creationFilter2 = new NotificationFilterSupport();
        creationFilter2.enableType("JMX.mbean.registered");
        creationFilter = creationFilter2;
        logger = new ClassLogger("javax.management.remote.misc", "ArrayNotificationBuffer");
        broadcasterClass = NotificationBroadcaster.class.getName();
    }
    
    private class ShareBuffer implements NotificationBuffer
    {
        private final int size;
        
        ShareBuffer(int size) {
            this.size = size;
            ArrayNotificationBuffer.this.addSharer(this);
        }
        
        @Override
        public NotificationResult fetchNotifications(NotificationBufferFilter notificationBufferFilter, long n, long n2, int n3) throws InterruptedException {
            return ArrayNotificationBuffer.this.fetchNotifications(notificationBufferFilter, n, n2, n3);
        }
        
        @Override
        public void dispose() {
            ArrayNotificationBuffer.access$000(ArrayNotificationBuffer.this, this);
        }
        
        int getSize() {
            return this.size;
        }
    }
    
    private static class NamedNotification
    {
        private final ObjectName sender;
        private final Notification notification;
        
        NamedNotification(ObjectName sender, Notification notification) {
            this.sender = sender;
            this.notification = notification;
        }
        
        ObjectName getObjectName() {
            return this.sender;
        }
        
        Notification getNotification() {
            return this.notification;
        }
        
        @Override
        public String toString() {
            return "NamedNotification(" + this.sender + ", " + this.notification + ")";
        }
    }
    
    private class BufferListener implements NotificationListener
    {
        @Override
        public void handleNotification(Notification obj, Object obj2) {
            if (ArrayNotificationBuffer.logger.debugOn()) {
                ArrayNotificationBuffer.logger.debug("BufferListener.handleNotification", "notif=" + obj + "; handback=" + obj2);
            }
            ArrayNotificationBuffer.this.addNotification(new ArrayNotificationBuffer.NamedNotification((ObjectName)obj2, obj));
        }
    }
    
    private static class BroadcasterQuery extends QueryEval implements QueryExp
    {
        private static final long serialVersionUID = 7378487660587592048L;
        
        @Override
        public boolean apply(ObjectName objectName) {
            return ArrayNotificationBuffer.access$400(QueryEval.getMBeanServer(), objectName, ArrayNotificationBuffer.broadcasterClass);
        }
    }
}
